from __future__ import print_function

import numpy as np

from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.tests.helper import assert_quantity_allclose

from ...core import PixCoord
from ...tests.helpers import make_simple_wcs
from ..annulus import *
from .test_common import BaseTestPixelRegion, BaseTestSkyRegion
from .utils import ASTROPY_LT_13


class TestCircleAnnulusPixelRegion(BaseTestPixelRegion):
    reg = CircleAnnulusPixelRegion(PixCoord(3, 4), inner_radius=2, outer_radius=3)
    sample_box = [0, 6, 1, 7]
    inside = [(3, 2)]
    outside = [(3, 0)]
    expected_area = 5 * np.pi
    expected_repr = '<CircleAnnulusPixelRegion(PixCoord(x=3, y=4), inner radius=2, outer radius=3)>'
    expected_str = 'Region: CircleAnnulusPixelRegion\ncenter: PixCoord(x=3, y=4)\ninner radius: 2\nouter radius: 3'

    skycoord = SkyCoord(3 * u.deg, 4 * u.deg, frame='icrs')
    wcs = make_simple_wcs(skycoord, 5 * u.arcsec, 20)

    def test_init(self):
        assert_quantity_allclose(self.reg.center.x, 3)
        assert_quantity_allclose(self.reg.center.y, 4)
        assert_quantity_allclose(self.reg.inner_radius, 2)
        assert_quantity_allclose(self.reg.outer_radius, 3)

    def test_transformation(self):
        skyannulus = self.reg.to_sky(wcs=self.wcs)
        assert isinstance(skyannulus, CircleAnnulusSkyRegion)


class TestCircleAnnulusSkyRegion(BaseTestSkyRegion):

    reg = CircleAnnulusSkyRegion(SkyCoord(3 * u.deg, 4 * u.deg), 20 * u.arcsec, 30 * u.arcsec)
    skycoord = SkyCoord(3 * u.deg, 4 * u.deg, frame='icrs')
    wcs = make_simple_wcs(skycoord, 5 * u.arcsec, 20)

    if ASTROPY_LT_13:
        expected_repr = ('<CircleAnnulusSkyRegion(<SkyCoord (ICRS): (ra, dec) in '
                         'deg\n    (3.0, 4.0)>, inner radius=20.0 arcsec, outer radius=30.0 arcsec)>')
        expected_str = ('Region: CircleAnnulusSkyRegion\ncenter: <SkyCoord (ICRS): '
                        '(ra, dec) in deg\n    (3.0, 4.0)>\ninner radius: 20.0 '
                        'arcsec\nouter radius: 30.0 arcsec')
    else:
        expected_repr = ('<CircleAnnulusSkyRegion(<SkyCoord (ICRS): (ra, dec) in '
                         'deg\n    ( 3.,  4.)>, inner radius=20.0 arcsec, outer radius=30.0 arcsec)>')
        expected_str = ('Region: CircleAnnulusSkyRegion\ncenter: <SkyCoord (ICRS): '
                        '(ra, dec) in deg\n    ( 3.,  4.)>\ninner radius: 20.0 '
                        'arcsec\nouter radius: 30.0 arcsec')

    def test_init(self):
        assert_quantity_allclose(self.reg.center.ra, self.skycoord.ra)
        assert_quantity_allclose(self.reg.inner_radius, 20*u.arcsec)
        assert_quantity_allclose(self.reg.outer_radius, 30*u.arcsec)

    def test_contains(self):
        assert not self.reg.contains(self.skycoord, self.wcs)
        test_coord = SkyCoord(3 * u.deg, 10 * u.deg, frame='icrs')
        assert not self.reg.contains(test_coord, self.wcs)

    def test_transformation(self):
        pixannulus = self.reg.to_pixel(wcs=self.wcs)
        assert isinstance(pixannulus, CircleAnnulusPixelRegion)


class TestEllipseAnnulusPixelRegion(BaseTestPixelRegion):
    reg = EllipseAnnulusPixelRegion(PixCoord(3, 4), 2, 5, 5, 8)
    sample_box = [1, 6, 0, 9]
    inside = [(3, 7)]
    outside = [(3, 4)]
    expected_area = 7.5 * np.pi
    expected_repr = ('<EllipseAnnulusPixelRegion(PixCoord(x=3, y=4), '
                     'inner width=2, inner height=5, outer width=5, '
                     'outer height=8, angle=0.0 deg)>')
    expected_str =  ('Region: EllipseAnnulusPixelRegion\ncenter: PixCoord(x=3, y=4)'
                     '\ninner width: 2\ninner height: 5\nouter width: 5\n'
                     'outer height: 8\nangle: 0.0 deg')

    skycoord = SkyCoord(3 * u.deg, 4 * u.deg, frame='icrs')
    wcs = make_simple_wcs(skycoord, 5 * u.arcsec, 20)

    def test_init(self):
        assert_quantity_allclose(self.reg.center.x, 3)
        assert_quantity_allclose(self.reg.center.y, 4)
        assert_quantity_allclose(self.reg.inner_width, 2)
        assert_quantity_allclose(self.reg.inner_height, 5)
        assert_quantity_allclose(self.reg.outer_width, 5)
        assert_quantity_allclose(self.reg.outer_height, 8)

    def test_transformation(self):
        skyannulus = self.reg.to_sky(wcs=self.wcs)
        assert isinstance(skyannulus, EllipseAnnulusSkyRegion)


class TestEllipseAnnulusSkyRegion(BaseTestSkyRegion):

    skycoord = SkyCoord(3 * u.deg, 4 * u.deg, frame='icrs')
    reg = EllipseAnnulusSkyRegion(skycoord, 20 * u.arcsec, 50 * u.arcsec,
                                  50 * u.arcsec, 80 * u.arcsec)
    wcs = make_simple_wcs(skycoord, 5 * u.arcsec, 20)

    if ASTROPY_LT_13:
        expected_repr = ('<EllipseAnnulusSkyRegion(<SkyCoord (ICRS): (ra, dec) in '
                         'deg\n    (3.0, 4.0)>, inner width=20.0 arcsec, '
                         'inner height=50.0 arcsec, outer width=50.0 arcsec, '
                         'outer height=80.0 arcsec, angle=0.0 deg)>')
        expected_str = ('Region: EllipseAnnulusSkyRegion\ncenter: <SkyCoord (ICRS): '
                        '(ra, dec) in deg\n    (3.0, 4.0)>\ninner width: 20.0 arcsec\n'
                        'inner height: 50.0 arcsec\nouter width: 50.0 arcsec\n'
                        'outer height: 80.0 arcsec\nangle: 0.0 deg')
    else:
        expected_repr = ('<EllipseAnnulusSkyRegion(<SkyCoord (ICRS): (ra, dec) in '
                         'deg\n    (3., 4.)>, inner width=20.0 arcsec, '
                         'inner height=50.0 arcsec, outer width=50.0 arcsec, '
                         'outer height=80.0 arcsec, angle=0.0 deg)>')
        expected_str = ('Region: EllipseAnnulusSkyRegion\ncenter: <SkyCoord (ICRS): '
                        '(ra, dec) in deg\n    (3., 4.)>\ninner width: 20.0 arcsec\n'
                        'inner height: 50.0 arcsec\nouter width: 50.0 arcsec\n'
                        'outer height: 80.0 arcsec\nangle: 0.0 deg')

    def test_init(self):
        assert_quantity_allclose(self.reg.center.ra, self.skycoord.ra)
        assert_quantity_allclose(self.reg.inner_width, 20*u.arcsec)
        assert_quantity_allclose(self.reg.inner_height, 50*u.arcsec)

    def test_contains(self):
        assert not self.reg.contains(self.skycoord, self.wcs)
        test_coord = SkyCoord(3 * u.deg, 10 * u.deg, frame='icrs')
        assert not self.reg.contains(test_coord, self.wcs)

    def test_transformation(self):
        pixannulus = self.reg.to_pixel(wcs=self.wcs)
        assert isinstance(pixannulus, EllipseAnnulusPixelRegion)


class TestRectangleAnnulusPixelRegion(BaseTestPixelRegion):
    reg = RectangleAnnulusPixelRegion(PixCoord(3, 4), 2, 5, 5, 8)
    sample_box = [1, 6, 0, 9]
    inside = [(3, 7)]
    outside = [(3, 4)]
    expected_area = 30
    expected_repr = ('<RectangleAnnulusPixelRegion(PixCoord(x=3, y=4), '
                     'inner width=2, inner height=5, outer width=5, '
                     'outer height=8, angle=0.0 deg)>')
    expected_str =  ('Region: RectangleAnnulusPixelRegion\ncenter: PixCoord(x=3, y=4)'
                     '\ninner width: 2\ninner height: 5\nouter width: 5\n'
                     'outer height: 8\nangle: 0.0 deg')

    skycoord = SkyCoord(3 * u.deg, 4 * u.deg, frame='icrs')
    wcs = make_simple_wcs(skycoord, 5 * u.arcsec, 20)

    def test_init(self):
        assert_quantity_allclose(self.reg.center.x, 3)
        assert_quantity_allclose(self.reg.center.y, 4)
        assert_quantity_allclose(self.reg.inner_width, 2)
        assert_quantity_allclose(self.reg.inner_height, 5)
        assert_quantity_allclose(self.reg.outer_width, 5)
        assert_quantity_allclose(self.reg.outer_height, 8)

    def test_transformation(self):
        skyannulus = self.reg.to_sky(wcs=self.wcs)
        assert isinstance(skyannulus, RectangleAnnulusSkyRegion)


class TestRectangleAnnulusSkyRegion(BaseTestSkyRegion):

    skycoord = SkyCoord(3 * u.deg, 4 * u.deg, frame='icrs')
    reg = RectangleAnnulusSkyRegion(skycoord, 20 * u.arcsec, 50 * u.arcsec,
                                  50 * u.arcsec, 80 * u.arcsec)
    wcs = make_simple_wcs(skycoord, 5 * u.arcsec, 20)

    if ASTROPY_LT_13:
        expected_repr = ('<RectangleAnnulusSkyRegion(<SkyCoord (ICRS): (ra, dec) in '
                         'deg\n    (3.0, 4.0)>, inner width=20.0 arcsec, '
                         'inner height=50.0 arcsec, outer width=50.0 arcsec, '
                         'outer height=80.0 arcsec, angle=0.0 deg)>')
        expected_str = ('Region: RectangleAnnulusSkyRegion\ncenter: <SkyCoord (ICRS): '
                        '(ra, dec) in deg\n    (3.0, 4.0)>\ninner width: 20.0 arcsec\n'
                        'inner height: 50.0 arcsec\nouter width: 50.0 arcsec\n'
                        'outer height: 80.0 arcsec\nangle: 0.0 deg')
    else:
        expected_repr = ('<RectangleAnnulusSkyRegion(<SkyCoord (ICRS): (ra, dec) in '
                         'deg\n    (3., 4.)>, inner width=20.0 arcsec, '
                         'inner height=50.0 arcsec, outer width=50.0 arcsec, '
                         'outer height=80.0 arcsec, angle=0.0 deg)>')
        expected_str = ('Region: RectangleAnnulusSkyRegion\ncenter: <SkyCoord (ICRS): '
                        '(ra, dec) in deg\n    (3., 4.)>\ninner width: 20.0 arcsec\n'
                        'inner height: 50.0 arcsec\nouter width: 50.0 arcsec\n'
                        'outer height: 80.0 arcsec\nangle: 0.0 deg')

    def test_init(self):
        assert_quantity_allclose(self.reg.center.ra, self.skycoord.ra)
        assert_quantity_allclose(self.reg.inner_width, 20*u.arcsec)
        assert_quantity_allclose(self.reg.inner_height, 50*u.arcsec)

    def test_contains(self):
        assert not self.reg.contains(self.skycoord, self.wcs)
        test_coord = SkyCoord(3 * u.deg, 10 * u.deg, frame='icrs')
        assert not self.reg.contains(test_coord, self.wcs)

    def test_transformation(self):
        pixannulus = self.reg.to_pixel(wcs=self.wcs)
        assert isinstance(pixannulus, RectangleAnnulusPixelRegion)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals

import operator
import six
import abc


from astropy import units as u
from astropy.wcs.utils import pixel_to_skycoord

from .._utils.wcs_helpers import skycoord_to_pixel_scale_angle
from ..core import CompoundPixelRegion, CompoundSkyRegion, PixCoord
from ..shapes.circle import CirclePixelRegion, CircleSkyRegion
from ..shapes.ellipse import EllipsePixelRegion, EllipseSkyRegion
from ..shapes.rectangle import RectanglePixelRegion, RectangleSkyRegion
from ..core.attributes import (AnnulusCenterPix, AnnulusInnerScalarLength,
                         AnnulusOuterScalarLength, AnnulusCenterSky,
                         AnnulusInnerQuantityLength, AnnulusOuterQuantityLength,
                         AnnulusAngle)

__all__ = ['CircleAnnulusPixelRegion', 'CircleAnnulusSkyRegion',
           'EllipseAnnulusPixelRegion', 'EllipseAnnulusSkyRegion',
           'RectangleAnnulusPixelRegion', 'RectangleAnnulusSkyRegion'
           ]


class CircleAnnulusPixelRegion(CompoundPixelRegion):
    """
    A circular annulus in pixel coordinates.

    Parameters
    ----------
    center : :class:`~regions.core.pixcoord.PixCoord`
        The position of the center of the annulus.
    inner_radius : `float`
        The inner radius of the annulus
    outer_radius : `float`
        The outer radius of the annulus
    meta : `~regions.RegionMeta` object, optional
        A dictionary which stores the meta attributes of this region.
    visual : `~regions.RegionVisual` object, optional
        A dictionary which stores the visual meta attributes of this region.

    Examples
    --------

    .. plot::
        :include-source:

        from regions import PixCoord, CircleAnnulusPixelRegion
        import matplotlib.pyplot as plt

        x, y = 6, 6
        inner_radius = 5.5
        outer_radius = 8
        fig, ax = plt.subplots(1, 1)

        center = PixCoord(x=x, y=y)
        reg = CircleAnnulusPixelRegion(center=center, inner_radius=inner_radius,
                                       outer_radius=outer_radius)
        patch = reg.as_artist(facecolor='none', edgecolor='red', lw=2)
        ax.add_patch(patch)

        plt.xlim(-5, 20)
        plt.ylim(-5, 20)
        ax.set_aspect('equal')
        plt.show()
    """

    center = AnnulusCenterPix()
    inner_radius = AnnulusInnerScalarLength("radius")
    outer_radius = AnnulusOuterScalarLength("radius")

    def __init__(self, center, inner_radius, outer_radius, meta=None, visual=None):
        region1 = CirclePixelRegion(center, inner_radius)
        region2 = CirclePixelRegion(center, outer_radius)
        super(CircleAnnulusPixelRegion, self).__init__(region1=region1,
                                                       region2=region2,
                                                       operator=operator.xor,
                                                       meta=meta,
                                                       visual=visual)
        self._repr_params = ('inner_radius', 'outer_radius')

    @property
    def area(self):
        return self.region2.area - self.region1.area

    @property
    def bounding_box(self):
        return self.region2.bounding_box

    def to_sky(self, wcs):
        center = pixel_to_skycoord(self.center.x, self.center.y, wcs)
        _, scale, _ = skycoord_to_pixel_scale_angle(center, wcs)
        inner_radius = self.inner_radius / scale * u.deg
        outer_radius = self.outer_radius / scale * u.deg
        return CircleAnnulusSkyRegion(center, inner_radius, outer_radius,
                                      self.meta, self.visual)


class CircleAnnulusSkyRegion(CompoundSkyRegion):
    """
    A circular annulus in sky coordinates.

    Parameters
    ----------
    center : :class:`~astropy.coordinates.SkyCoord`
        The position of the center of the annulus.
    inner_radius : :class:`~astropy.units.Quantity`
        The inner radius of the annulus in angular units
    outer_radius : :class:`~astropy.units.Quantity`
        The outer radius of the annulus in angular units
    meta : `~regions.RegionMeta` object, optional
        A dictionary which stores the meta attributes of this region.
    visual : `~regions.RegionVisual` object, optional
        A dictionary which stores the visual meta attributes of this region.
    """

    center = AnnulusCenterSky()
    inner_radius = AnnulusInnerQuantityLength("radius")
    outer_radius = AnnulusOuterQuantityLength("radius")

    def __init__(self, center, inner_radius, outer_radius, meta=None, visual=None):
        region1 = CircleSkyRegion(center, inner_radius)
        region2 = CircleSkyRegion(center, outer_radius)
        super(CircleAnnulusSkyRegion, self).__init__(region1=region1,
                                                     operator=operator.xor,
                                                     region2=region2,
                                                     meta=meta,
                                                     visual=visual)
        self._repr_params = ('inner_radius', 'outer_radius')

    def to_pixel(self, wcs):
        center, scale, _ = skycoord_to_pixel_scale_angle(self.center, wcs)
        # FIXME: The following line is needed to get a scalar PixCoord
        center = PixCoord(float(center.x), float(center.y))
        inner_radius = self.inner_radius.to('deg').value * scale
        outer_radius = self.outer_radius.to('deg').value * scale
        return CircleAnnulusPixelRegion(center, inner_radius, outer_radius,
                                        self.meta, self.visual)


@six.add_metaclass(abc.ABCMeta)
class AsymmetricAnnulusPixelRegion(CompoundPixelRegion):
    """
    An asymmetrical annulus in pixel coordinates.

    region1 : `~regions.PixelRegion` object
        The inner asymmetric region
    region2 : `~regions.PixelRegion` object
        The outer asymmetric region
    meta : `~regions.RegionMeta` object, optional
        A dictionary which stores the meta attributes of this region.
    visual : `~regions.RegionVisual` object, optional
        A dictionary which stores the visual meta attributes of this region.
    """

    center = AnnulusCenterPix()
    inner_width = AnnulusInnerScalarLength('width')
    outer_width = AnnulusOuterScalarLength('width')
    inner_height = AnnulusInnerScalarLength('height')
    outer_height = AnnulusOuterScalarLength('height')
    angle = AnnulusAngle()

    def __init__(self, region1, region2, meta=None, visual=None):

        super(AsymmetricAnnulusPixelRegion, self).__init__(region1=region1,
                                                           region2=region2,
                                                           operator=operator.xor,
                                                           meta=meta,
                                                           visual=visual)

        self._repr_params = ('inner_width', 'inner_height',
                             'outer_width', 'outer_height', 'angle')

    @property
    def area(self):
        return self.region2.area - self.region1.area

    @property
    def bounding_box(self):
        return self.region2.bounding_box

    def to_sky_args(self, wcs):

        center = pixel_to_skycoord(self.center.x, self.center.y, wcs)
        _, scale, north_angle = skycoord_to_pixel_scale_angle(center, wcs)

        inner_width = self.inner_width / scale * u.deg
        inner_height = self.inner_height / scale * u.deg
        outer_width = self.outer_width / scale * u.deg
        outer_height = self.outer_height / scale * u.deg
        angle = self.angle - (north_angle - 90 * u.deg)

        return center, inner_width, inner_height, outer_width, outer_height, angle


@six.add_metaclass(abc.ABCMeta)
class AsymmetricAnnulusSkyRegion(CompoundSkyRegion):
    """
    A rectangular annulus in `~astropy.coordinates.SkyCoord` coordinates.

    Parameters
    ----------
    region1 : `~regions.SkyRegion` object
        The inner asymmetric region
    region2 : `~regions.SkyRegion` object
        The outer asymmetric region
    meta : `~regions.RegionMeta` object, optional
        A dictionary which stores the meta attributes of this region.
    visual : `~regions.RegionVisual` object, optional
        A dictionary which stores the visual meta attributes of this region.
    """

    center = AnnulusCenterSky()
    inner_width = AnnulusInnerQuantityLength('width')
    outer_width = AnnulusOuterQuantityLength('width')
    inner_height = AnnulusInnerQuantityLength('height')
    outer_height = AnnulusOuterQuantityLength('height')
    angle = AnnulusAngle()

    def __init__(self, region1, region2, meta=None, visual=None):

        super(AsymmetricAnnulusSkyRegion, self).__init__(region1=region1,
                                                         region2=region2,
                                                         operator=operator.xor,
                                                         meta=meta,
                                                         visual=visual)

        self._repr_params = ('inner_width', 'inner_height',
                             'outer_width', 'outer_height', 'angle')

    def to_pixel_args(self, wcs):
        center, scale, north_angle = skycoord_to_pixel_scale_angle(self.center,
                                                                   wcs)
        center = PixCoord(center.x, center.y)
        inner_width = self.inner_width.to('deg').value * scale
        inner_height = self.inner_height.to('deg').value * scale
        outer_width = self.outer_width.to('deg').value * scale
        outer_height = self.outer_height.to('deg').value * scale
        angle = self.angle + (north_angle - 90 * u.deg)

        return center, inner_width, inner_height, outer_width, outer_height, angle


class EllipseAnnulusPixelRegion(AsymmetricAnnulusPixelRegion):
    """
    A elliptical annulus in pixel coordinates.

    Parameters
    ----------
    center : `~regions.PixCoord`
        The position of the center of the elliptical annulus.
    inner_width : `float`
        The inner width of the elliptical annulus (before rotation) in pixels
    inner_height : `float`
        The inner height of the elliptical annulus (before rotation) in pixels
    outer_width : `float`
        The outer width of the elliptical annulus (before rotation) in pixels
    outer_height : `float`
        The outer height of the elliptical annulus (before rotation) in pixels
    angle : `~astropy.units.Quantity`, optional
        The rotation angle of the elliptical annulus, measured anti-clockwise.
        If set to zero (the default), the width axis is lined up with the x axis.
    meta : `~regions.RegionMeta` object, optional
        A dictionary which stores the meta attributes of this region.
    visual : `~regions.RegionVisual` object, optional
        A dictionary which stores the visual meta attributes of this region.


    Examples
    --------

    .. plot::
        :include-source:

        from astropy.coordinates import Angle
        from regions import PixCoord, EllipseAnnulusPixelRegion
        import matplotlib.pyplot as plt

        x, y = 6, 6
        inner_width = 5.5
        inner_height = 3.5
        outer_width = 8.5
        outer_height = 6.5
        angle = Angle("45deg")

        fig, ax = plt.subplots(1, 1)

        center = PixCoord(x=x, y=y)
        reg = EllipseAnnulusPixelRegion(center=center, inner_width=inner_width,
                                        inner_height=inner_height, outer_width=outer_width,
                                        outer_height=outer_height, angle=angle)
        patch = reg.as_artist(facecolor='none', edgecolor='red', lw=2)
        ax.add_patch(patch)

        plt.xlim(-5, 20)
        plt.ylim(-5, 20)
        ax.set_aspect('equal')
        plt.show()
    """

    def __init__(self, center, inner_width, inner_height, outer_width,
                 outer_height, angle=0 * u.deg, meta=None, visual=None):
        region1 = EllipsePixelRegion(center, inner_width, inner_height, angle)
        region2 = EllipsePixelRegion(center, outer_width, outer_height, angle)

        super(EllipseAnnulusPixelRegion, self).__init__(region1=region1,
                                                        region2=region2,
                                                        meta=meta,
                                                        visual=visual)

    def to_sky(self, wcs):
        return EllipseAnnulusSkyRegion(*self.to_sky_args(wcs),
                                       meta=self.meta, visual=self.visual)


class EllipseAnnulusSkyRegion(AsymmetricAnnulusSkyRegion):
    """
    A elliptical annulus in `~astropy.coordinates.SkyCoord` coordinates.

    Parameters
    ----------
    center : `~astropy.coordinates.SkyCoord`
        The position of the center of the elliptical annulus.
    inner_width : `~astropy.units.Quantity`
        The inner width of the elliptical annulus (before rotation) as angle
    inner_height : `~astropy.units.Quantity`
        The inner height of the elliptical annulus (before rotation) as angle
    outer_width : `~astropy.units.Quantity`
        The outer width of the elliptical annulus (before rotation) as angle
    outer_height : `~astropy.units.Quantity`
        The outer height of the elliptical annulus (before rotation) as angle
    angle : `~astropy.units.Quantity`, optional
        The rotation angle of the elliptical annulus, measured anti-clockwise.
        If set to zero (the default), the width axis is lined up with the
        longitude axis of the celestial coordinates
    meta : `~regions.RegionMeta` object, optional
        A dictionary which stores the meta attributes of this region.
    visual : `~regions.RegionVisual` object, optional
        A dictionary which stores the visual meta attributes of this region.
    """

    def __init__(self, center, inner_width, inner_height, outer_width,
                 outer_height, angle=0 * u.deg, meta=None, visual=None):
        region1 = EllipseSkyRegion(center, inner_width, inner_height, angle)
        region2 = EllipseSkyRegion(center, outer_width, outer_height, angle)

        super(EllipseAnnulusSkyRegion, self).__init__(region1=region1,
                                                      region2=region2,
                                                      meta=meta,
                                                      visual=visual)

    def to_pixel(self, wcs):
        return EllipseAnnulusPixelRegion(*self.to_pixel_args(wcs),
                                         meta=self.meta, visual=self.visual)


class RectangleAnnulusPixelRegion(AsymmetricAnnulusPixelRegion):
    """
    A rectangular annulus in pixel coordinates.

    Parameters
    ----------
    center : `~regions.PixCoord`
        The position of the center of the rectangular annulus.
    inner_width : `float`
        The inner width of the rectangular annulus (before rotation) in pixels
    inner_height : `float`
        The inner height of the rectangular annulus (before rotation) in pixels
    outer_width : `float`
        The outer width of the rectangular annulus (before rotation) in pixels
    outer_height : `float`
        The outer height of the rectangular annulus (before rotation) in pixels
    angle : `~astropy.units.Quantity`, optional
        The rotation angle of the rectangular annulus, measured anti-clockwise.
        If set to zero (the default), the width axis is lined up with the x axis.
    meta : `~regions.RegionMeta` object, optional
        A dictionary which stores the meta attributes of this region.
    visual : `~regions.RegionVisual` object, optional
        A dictionary which stores the visual meta attributes of this region.


    Examples
    --------

    .. plot::
        :include-source:

        from astropy.coordinates import Angle
        from regions import PixCoord, RectangleAnnulusPixelRegion
        import matplotlib.pyplot as plt

        x, y = 6, 6
        inner_width = 5.5
        inner_height = 3.5
        outer_width = 8.5
        outer_height = 6.5
        angle = Angle("45deg")

        fig, ax = plt.subplots(1, 1)

        center = PixCoord(x=x, y=y)
        reg = RectangleAnnulusPixelRegion(center=center, inner_width=inner_width,
                                          inner_height=inner_height, outer_width=outer_width,
                                          outer_height=outer_height, angle=angle)
        patch = reg.as_artist(facecolor='none', edgecolor='red', lw=2)
        ax.add_patch(patch)

        plt.xlim(-5, 20)
        plt.ylim(-5, 20)
        ax.set_aspect('equal')
        plt.show()
    """

    def __init__(self, center, inner_width, inner_height, outer_width,
                 outer_height, angle=0 * u.deg, meta=None, visual=None):
        region1 = RectanglePixelRegion(center, inner_width, inner_height, angle)
        region2 = RectanglePixelRegion(center, outer_width, outer_height, angle)

        super(RectangleAnnulusPixelRegion, self).__init__(
            region1=region1, region2=region2, meta=meta, visual=visual)

    def to_sky(self, wcs):
        return RectangleAnnulusSkyRegion(*self.to_sky_args(wcs),
                                         meta=self.meta, visual=self.visual)


class RectangleAnnulusSkyRegion(AsymmetricAnnulusSkyRegion):
    """
    A rectangular annulus in `~astropy.coordinates.SkyCoord` coordinates.

    Parameters
    ----------
    center : `~astropy.coordinates.SkyCoord`
        The position of the center of the rectangular annulus.
    inner_width : `~astropy.units.Quantity`
        The inner width of the rectangular annulus (before rotation) as angle
    inner_height : `~astropy.units.Quantity`
        The inner height of the rectangular annulus (before rotation) as angle
    outer_width : `~astropy.units.Quantity`
        The outer width of the rectangular annulus (before rotation) as angle
    outer_height : `~astropy.units.Quantity`
        The outer height of the rectangular annulus (before rotation) as angle
    angle : `~astropy.units.Quantity`, optional
        The rotation angle of the rectangular annulus, measured anti-clockwise.
        If set to zero (the default), the width axis is lined up with the
        longitude axis of the celestial coordinates
    meta : `~regions.RegionMeta` object, optional
        A dictionary which stores the meta attributes of this region.
    visual : `~regions.RegionVisual` object, optional
        A dictionary which stores the visual meta attributes of this region.
    """

    def __init__(self, center, inner_width, inner_height, outer_width,
                 outer_height, angle=0 * u.deg, meta=None, visual=None):
        region1 = RectangleSkyRegion(center, inner_width, inner_height, angle)
        region2 = RectangleSkyRegion(center, outer_width, outer_height, angle)

        super(RectangleAnnulusSkyRegion, self).__init__(
            region1=region1, region2=region2, meta=meta, visual=visual)

    def to_pixel(self, wcs):
        return RectangleAnnulusPixelRegion(*self.to_pixel_args(wcs),
                                           meta=self.meta, visual=self.visual)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from distutils.version import LooseVersion
import astropy

ASTROPY_LT_13 = LooseVersion(astropy.__version__) < LooseVersion('1.3')

try:
    import matplotlib  # noqa
    HAS_MATPLOTLIB = True
except:
    HAS_MATPLOTLIB = False



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np
from numpy.testing import assert_allclose

import astropy.units as u
from astropy.coordinates import SkyCoord

from ...shapes import CircleSkyRegion, CirclePixelRegion
from ...core import PixCoord, CompoundPixelRegion
from ...tests.helpers import make_simple_wcs


def test_compound_pixel():
    # Two circles that overlap in one column
    pixcoord1 = PixCoord(5, 5)
    c1 = CirclePixelRegion(pixcoord1, 4)
    pixcoord2 = PixCoord(11, 5)
    c2 = CirclePixelRegion(pixcoord2, 4)

    union = c1 | c2
    assert isinstance(union, CompoundPixelRegion)
    mask = union.to_mask()
    assert_allclose(mask.data[:,7], [0, 0, 1, 1, 1, 1, 1, 0, 0])
    assert_allclose(mask.data[:,6], [0, 1, 1, 1, 1, 1, 1, 1, 0])

    intersection = c1 & c2
    mask = intersection.to_mask()
    assert_allclose(mask.data[:,7], [0, 0, 1, 1, 1, 1, 1, 0, 0])
    assert_allclose(mask.data[:,6], [0, 0, 0, 0, 0, 0, 0, 0, 0])

    symdiff = c1 ^ c2
    mask = symdiff.to_mask()
    assert_allclose(mask.data[:,7], [0, 0, 0, 0, 0, 0, 0, 0, 0])
    assert_allclose(mask.data[:,6], [0, 1, 1, 1, 1, 1, 1, 1, 0])

    # fixes #168
    pixcoord3 = PixCoord(1, 1)
    c3 = CirclePixelRegion(pixcoord3, 4)
    union2 = c1 | c3
    mask1 = union2.to_mask()

    pixcoord4 = PixCoord(9, 9)
    c4 = CirclePixelRegion(pixcoord4, 4)
    union3 = c1 | c4
    mask2 = union3.to_mask()

    # mask1 and mask2 should be equal
    assert_allclose(mask1.data, mask2.data)

    ref_data = np.array([[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                         [0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0],
                         [0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0],
                         [0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0],
                         [0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0],
                         [0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0],
                         [0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0],
                         [0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0],
                         [0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0],
                         [0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0],
                         [0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0],
                         [0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0],
                         [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]]
                        )
    assert_allclose(mask1.data, ref_data)


def test_compound_sky():

    skycoord1 = SkyCoord(0 * u.deg, 0 * u.deg, frame='galactic')
    c1 = CircleSkyRegion(skycoord1, 1 * u.deg)

    skycoord2 = SkyCoord(1 * u.deg, 1 * u.deg, frame='galactic')
    c2 = CircleSkyRegion(skycoord2, 0.5 * u.deg)

    test_coord1 = SkyCoord(1.2 * u.deg, 1.2 * u.deg, frame='galactic')
    test_coord2 = SkyCoord(0.5 * u.deg, 0.5 * u.deg, frame='galactic')
    test_coord3 = SkyCoord(0.7 * u.deg, 0.7 * u.deg, frame='galactic')
    test_coord4 = SkyCoord(2 * u.deg, 5 * u.deg, frame='galactic')

    wcs = make_simple_wcs(skycoord1, 0.1 * u.deg, 20)

    assert c2.contains(test_coord1, wcs) and not c1.contains(test_coord1, wcs)
    assert not c2.contains(test_coord2, wcs) and c1.contains(test_coord2, wcs)
    assert c1.contains(test_coord3, wcs) and c2.contains(test_coord3, wcs)
    assert not c2.contains(test_coord4, wcs) and not c1.contains(test_coord4, wcs)

    coords = SkyCoord([test_coord1, test_coord2, test_coord3, test_coord4], frame='galactic')

    union = c1 | c2
    assert (union.contains(coords, wcs) == [True, True, True, False]).all()

    intersection = c1 & c2
    assert (intersection.contains(coords, wcs) == [False, False, True, False]).all()

    diff = c1 ^ c2
    assert (diff.contains(coords, wcs) == [True, True, False, False]).all()

    c3 = CircleSkyRegion(test_coord4, 0.1 * u.deg)

    union = c1 | c2 | c3
    assert (union.contains(coords, wcs) == [True, True, True, True]).all()

    intersection = c1 & c2 & c3
    assert (intersection.contains(coords, wcs) == [False, False, False, False]).all()

    diff = c1 ^ c2 ^ c3
    assert (diff.contains(coords, wcs) == [True, True, False, True]).all()

    assert 'Compound' in str(union)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import numpy as np
import astropy.units as u


__all__ = ['RegionMask']


class RegionMask(object):
    """
    Class for a region mask.

    Parameters
    ----------
    data : array_like
        A 2D array of a region mask representing the fractional overlap
        of the region on the pixel grid.  This should be the full-sized
        (i.e. not truncated) array that is the direct output of one of
        the low-level "geometry" functions.

    bbox : `regions.BoundingBox`
        The bounding box for the region.

    Examples
    --------
    Usage examples are provided in the :ref:`gs-masks` section of the
    docs.
    """

    def __init__(self, data, bbox):
        self.data = np.asanyarray(data)
        if self.data.shape != bbox.shape:
            raise ValueError('data and bounding box must have the same '
                             'shape.')
        self.bbox = bbox

    def __array__(self):
        """
        Array representation of the mask data array (e.g., for
        matplotlib).
        """

        return self.data

    @property
    def shape(self):
        """
        The shape of the mask data array.
        """

        return self.data.shape

    def _overlap_slices(self, shape):
        """
        Calculate the slices for the overlapping part of the bounding
        box and an array of the given shape.

        Parameters
        ----------
        shape : tuple of int
            The ``(ny, nx)`` shape of array where the slices are to be
            applied.

        Returns
        -------
        slices_large : tuple of slices
            A tuple of slice objects for each axis of the large array,
            such that ``large_array[slices_large]`` extracts the region
            of the large array that overlaps with the small array.

        slices_small : slice
            A tuple of slice objects for each axis of the small array,
            such that ``small_array[slices_small]`` extracts the region
            of the small array that is inside the large array.
        """

        if len(shape) != 2:
            raise ValueError('input shape must have 2 elements.')

        xmin = self.bbox.ixmin
        xmax = self.bbox.ixmax
        ymin = self.bbox.iymin
        ymax = self.bbox.iymax

        if xmin >= shape[1] or ymin >= shape[0] or xmax <= 0 or ymax <= 0:
            # no overlap of the region with the data
            return None, None

        slices_large = (slice(max(ymin, 0), min(ymax, shape[0])),
                        slice(max(xmin, 0), min(xmax, shape[1])))

        slices_small = (slice(max(-ymin, 0),
                              min(ymax - ymin, shape[0] - ymin)),
                        slice(max(-xmin, 0),
                              min(xmax - xmin, shape[1] - xmin)))

        return slices_large, slices_small

    def _to_image_partial_overlap(self, image):
        """
        Return an image of the mask in a 2D array, where the mask
        is not fully within the image (i.e. partial or no overlap).
        """

        # find the overlap of the mask on the output image shape
        slices_large, slices_small = self._overlap_slices(image.shape)

        if slices_small is None:
            return None    # no overlap

        # insert the mask into the output image
        image[slices_large] = self.data[slices_small]

        return image

    def to_image(self, shape):
        """
        Return an image of the mask in a 2D array of the given shape,
        taking any edge effects into account.

        Parameters
        ----------
        shape : tuple of int
            The ``(ny, nx)`` shape of the output array.

        Returns
        -------
        result : `~numpy.ndarray`
            A 2D array of the mask.
        """

        if len(shape) != 2:
            raise ValueError('input shape must have 2 elements.')

        image = np.zeros(shape)

        if self.bbox.ixmin < 0 or self.bbox.iymin < 0:
            return self._to_image_partial_overlap(image)

        try:
            image[self.bbox.slices] = self.data
        except ValueError:    # partial or no overlap
            image = self._to_image_partial_overlap(image)

        return image

    def cutout(self, data, fill_value=0., copy=False):
        """
        Create a cutout from the input data over the mask bounding box,
        taking any edge effects into account.

        Parameters
        ----------
        data : array_like
            A 2D array on which to apply the region mask.

        fill_value : float, optional
            The value is used to fill pixels where the region mask
            does not overlap with the input ``data``.  The default is 0.

        copy : bool, optional
            If `True` then the returned cutout array will always be hold
            a copy of the input ``data``.  If `False` and the mask is
            fully within the input ``data``, then the returned cutout
            array will be a view into the input ``data``.  In cases
            where the mask partially overlaps or has no overlap with the
            input ``data``, the returned cutout array will always hold a
            copy of the input ``data`` (i.e. this keyword has no
            effect).

        Returns
        -------
        result : `~numpy.ndarray` or `None`
            A 2D array cut out from the input ``data`` representing the
            same cutout region as the region mask.  If there is a
            partial overlap of the region mask with the input data,
            pixels outside of the data will be assigned to
            ``fill_value``.  `None` is returned if there is no overlap
            of the region with the input ``data``.
        """

        data = np.asanyarray(data)
        if data.ndim != 2:
            raise ValueError('data must be a 2D array.')

        partial_overlap = False
        if self.bbox.ixmin < 0 or self.bbox.iymin < 0:
            partial_overlap = True

        if not partial_overlap:
            # try this for speed -- the result may still be a partial
            # overlap, in which case the next block will be triggered
            if copy:
                cutout = np.copy(data[self.bbox.slices])
            else:
                cutout = data[self.bbox.slices]

        if partial_overlap or (cutout.shape != self.shape):
            slices_large, slices_small = self._overlap_slices(data.shape)

            if slices_small is None:
                return None    # no overlap

            # cutout is a copy
            cutout = np.zeros(self.shape, dtype=data.dtype)
            cutout[:] = fill_value
            cutout[slices_small] = data[slices_large]

            if isinstance(data, u.Quantity):
                cutout = u.Quantity(cutout, unit=data.unit)

        return cutout

    def multiply(self, data, fill_value=0.):
        """
        Multiply the region mask with the input data, taking any edge
        effects into account.

        The result is a mask-weighted cutout from the data.

        Parameters
        ----------
        data : array_like or `~astropy.units.Quantity`
            The 2D array to multiply with the region mask.

        fill_value : float, optional
            The value is used to fill pixels where the region mask does
            not overlap with the input ``data``.  The default is 0.

        Returns
        -------
        result : `~numpy.ndarray` or `None`
            A 2D mask-weighted cutout from the input ``data``.  If there
            is a partial overlap of the region mask with the input data,
            pixels outside of the data will be assigned to
            ``fill_value`` before being multipled with the mask.  `None`
            is returned if there is no overlap of the region with the
            input ``data``.
        """

        cutout = self.cutout(data, fill_value=fill_value)
        if cutout is None:
            return None
        else:
            return cutout * self.data



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function
import re
import copy
import itertools
from warnings import warn

from astropy import units as u
from astropy import coordinates
from astropy.extern import six

from .core import CRTFRegionParserError, CRTFRegionParserWarning, valid_symbols
from ..core import Shape, ShapeList, reg_mapping

__all__ = ['read_crtf', 'CRTFParser', 'CRTFRegionParser']

# All CASA files start with '#CRTF' . It may also include the version number like '#CRTFv0' .
regex_begin = re.compile(r'^#CRTFv?[\d]?$')

# Comment Format :
regex_comment = re.compile(r'^#.*$')

# Identifies the global attributes Format.
regex_global = re.compile(r'^global\s+(?P<parameters>.*)?')

# Coordinate Format : "[x, y]"
regex_coordinate = re.compile(r'\[([\w.+-:]*?)\s*[,]\s*([\w.+-:]*?)\]')

# Single length Format. For Ex : helps us to extract the radius of a circle.
regex_length = re.compile(r'(?:\[[^=]*\])+[,]\s*([^\[]*)\]')

# Extracts each 'parameter=value' pair.
regex_meta = re.compile(r'(?:(\w+)\s*=[\s\'\"]*([^,\[\]]+?)[\'\",]+)|(?:(\w+)\s*=\s*\[(.*?)\])')

# Region format which segregates include('+'|'-') parameter, kind of definition ('ann' for annotations | '' for regions)
# and region type.
regex_region = re.compile(r'(?P<include>[+-])?(?P<type>ann(?=\s))?\s*(?P<regiontype>[a-z]*?)\[[^=]*]')

# Line format which checks the validity of the line and segregates the meta attributes from the region format.
regex_line = re.compile(r'(?P<region>[+-]?(?:ann(?=\s))?\s*[a-z]+?\[[^=]+\])(?:\s*[,]\s*(?P<parameters>.*))?')


def read_crtf(filename, errors='strict'):
    """
    Reads a CRTF region file and returns a list of region objects.

    Parameters
    ----------
    filename : `str`
        The file path
    errors : ``warn``, ``ignore``, ``strict``, optional
      The error handling scheme to use for handling parsing errors.
      The default is 'strict', which will raise a `~regions.CRTFRegionParserError`.
      ``warn`` will raise a `~regions.CRTFRegionParserWarning`, and ``ignore`` will do nothing
      (i.e., be silent).

    Returns
    -------
    regions : `list`
        Python `list` of `~regions.Region` objects.

    Examples
    --------
    >>> from regions import read_crtf
    >>> from astropy.utils.data import get_pkg_data_filename
    >>> file = get_pkg_data_filename('data/CRTFgeneral.crtf', package='regions.io.crtf.tests')
    >>> regs = read_crtf(file, errors='warn')
    >>> print(regs[0])
    Region: CircleSkyRegion
    center: <SkyCoord (FK4: equinox=B1950.000, obstime=B1950.000): (ra, dec) in deg
    (273.1, -23.18333333)>
    radius: 2.3 arcsec
    >>> print(regs[0].meta)
    {'frame': 'BARY', 'corr': ['I', 'Q'], 'include': True, 'type': 'ann'}
    >>> print(regs[0].visual)
    {'color': 'blue'}

    """
    with open(filename) as fh:
        if regex_begin.search(fh.readline()):
            region_string = fh.read()
            parser = CRTFParser(region_string, errors)
            return parser.shapes.to_regions()
        else:
            raise CRTFRegionParserError('Every CRTF Region must start with "#CRTF" ')


class CRTFParser(object):
    """
    Parses a CRTF string.

    This class transforms a CRTF string to a `~regions.io.core.ShapeList`. The
    result is stored as ``shapes`` attribute.

    Each line is tested for either containing a region with meta attributes or global parameters.
    If global parameters are found then, it is stored in the ``global_meta`` attribute.
    If a region is found the `~regions.CRTFRegionParser` is invoked to transform the line into a
    `~regions.io.core.Shape` object.

    Parameters
    ----------
    region_string : `str`
        CRTF region string
    errors : ``warn``, ``ignore``, ``strict``, optional
        The error handling scheme to use for handling parsing errors.
        The default is 'strict', which will raise a `~regions.CRTFRegionParserError`.
        ``warn`` will raise a `~regions.CRTFRegionParserWarning`, and ``ignore`` will do nothing
        (i.e., be silent).

    Examples
    --------
    >>> from regions import CRTFParser
    >>> reg_str = "ann circle[[18h12m24s, -23d11m00s], 2.3arcsec], coord=B1950, frame=BARY, corr=[I, Q], color=blue"
    >>> regs = CRTFParser(reg_str, errors='warn')
    >>> print(regs[0])
    Region: CircleSkyRegion
    center: <SkyCoord (FK4: equinox=B1950.000, obstime=B1950.000): (ra, dec) in deg
    (273.1, -23.18333333)>
    radius: 2.3 arcsec
    >>> print(regs[0].meta)
    {'frame': 'BARY', 'corr': ['I', 'Q'], 'include': True, 'type': 'ann'}
    >>> print(regs[0].visual)
    {'color': 'blue'}

    """

    # It contains a tuple of valid definition (region, annotation) type.
    valid_definition = ('box', 'centerbox', 'rotbox', 'poly', 'circle', 'annulus', 'ellipse',
                        'line', 'vector', 'text', 'symbol')

    # It contains a tuple of valid name of the parameters(attributes).
    valid_global_keys = ('coord', 'frame', 'corr', 'veltype', 'restfreq', 'linewidth', 'linestyle', 'symsize',
                         'symthick', 'color', 'font', 'fontsize', 'fontstyle', 'usetex', 'labelpos','labelcolor',
                         'labeloff', 'range')

    def __init__(self, region_string, errors='strict'):

        if errors not in ('strict', 'ignore', 'warn'):
            msg = "``errors`` must be one of strict, ignore, or warn; is {}"
            raise ValueError(msg.format(errors))

        self.region_string = region_string
        self.errors = errors

        # Global states
        self.global_meta = {}

        # Results
        self.shapes = ShapeList()

        self.run()

    def __str__(self):
        ss = self.__class__.__name__
        ss += '\nErrors: {}'.format(self.errors)
        ss += '\nGlobal meta: {}'.format(self.global_meta)
        ss += '\nShapes: {}'.format(self.shapes)
        ss += '\n'
        return ss

    def parse_line(self, line):
        """
        Parses a single line.
        """

        # Skip blanks
        if line == '':
            return

        # Skip comments
        if regex_comment.search(line):
            return

        # Special case / header: parse global parameters into metadata
        global_parameters = regex_global.search(line)
        if global_parameters:
            self.parse_global_meta(global_parameters.group('parameters'))
            return

        # Tries to check the validity of the line.
        crtf_line = regex_line.search(line)
        if crtf_line:
            # Tries to parse the line.

            # Finds info about the region.
            region = regex_region.search(crtf_line.group('region'))
            type_ = region.group('type') or 'reg'
            include = region.group('include') or '+'
            region_type = region.group('regiontype').lower()
            if region_type in self.valid_definition:
                helper = CRTFRegionParser(self.global_meta, include, type_, region_type,
                                          *crtf_line.group('region', 'parameters'))
                self.shapes.append(helper.shape)
            else:
                self._raise_error("Not a valid CRTF Region type: '{0}'.".format(region_type))

        else:
            self._raise_error("Not a valid CRTF line: '{0}'.".format(line))
            return

    def _raise_error(self, msg):
        if self.errors == 'warn':
            warn(msg, CRTFRegionParserWarning)
        elif self.errors == 'strict':
            raise CRTFRegionParserError(msg)

    def run(self):
        """
        Run all the steps.
        Splits the regions into line and calls ``parse_line`` for each line.
        """
        for line in self.region_string.split('\n'):
            self.parse_line(line)

    def parse_global_meta(self, global_meta_str):
        """
        Parses the line starting with global to extract all the valid meta key/value pair.
        """

        if global_meta_str:
            global_meta_str = regex_meta.findall(global_meta_str + ',')
        if global_meta_str:
            for par in global_meta_str:
                if par[0] is not '':
                    val1 = par[0].lower()
                    val2 = par[1]
                else:
                    val1 = par[2].lower()
                    val2 = par[3]
                val1 = val1.strip()
                val2 = val2.strip()
                if val1 in self.valid_global_keys :
                    if val1 in ('range', 'corr', 'labeloff'):
                        val2 = val2.split(",")
                        val2 = [x.strip() for x in val2 if x]
                    self.global_meta[val1] = val2
                else:
                    self._raise_error("'{0}' is not a valid global meta key".format(val1))


class CRTFRegionParser(object):
    """
    Parse a CRTF region string

    This will turn a line containing a CRTF region into a `~regions.Shape` object.

    Parameters
    ----------
    global_meta : `dict`
        Global meta data of the CRTF file which is used as default meta values for regions
    include : `str` {'+', '-'}
        Flag at the beginning of the line
    type_ : `str` {'reg', 'ann'}
        Kind of the region definition
    region_type : `str`
        Region type
    reg_str : `str`
        Region string to parse
    meta_str : `str`
        Meta string to parse
    errors : ``warn``, ``ignore``, ``strict``, optional
        The error handling scheme to use for handling parsing errors.
        The default is 'strict', which will raise a `~regions.CRTFRegionParserError`.
        ``warn`` will raise a `~regions.CRTFRegionParserWarning`, and
        ``ignore`` will do nothing (i.e., be silent).
    """

    # List of valid coordinate system
    # TODO : There are still many reference systems to support

    coordinate_systems = ['j2000', 'icrs', 'galactic', 'supergal', 'image', 'ecliptic']

    # Maps CASA coordinate frame to appropriate astropy coordinate frames.
    coordsys_mapping = dict(zip(coordinates.frame_transform_graph.get_names(),
                                coordinates.frame_transform_graph.get_names()))
    coordsys_mapping['j2000'] = 'fk5'
    coordsys_mapping['b1950'] = 'fk4'
    coordsys_mapping['supergal'] = 'supergalactic'
    coordsys_mapping['ecliptic'] = 'geocentrictrueecliptic'

    # CRTF Format specifications. This define how a certain region is read.
    # 'c' denotes a coordinates, 'l' denotes a length, 'pl' denotes a pair of lengths,
    # 's' denotes a string(generally a text or symbol)
    language_spec = {'circle': ['c', 'l'],
                     'box': ['c', 'c'],
                     'centerbox': ['c', 'pl'],
                     'rotbox': ['c', 'pl', 'l'],
                     'poly': itertools.cycle('c'),
                     'annulus': ['c', 'pl'],
                     'ellipse': ['c', 'pl', 'l'],
                     'line': ['c', 'c'],
                     'vector': ['c', 'c'],
                     'symbol': ['c', 's'],
                     'text': ['c', 's']}

    def __init__(self, global_meta, include, type_, region_type, reg_str, meta_str, errors='strict'):

        self.global_meta = global_meta
        self.reg_str = reg_str
        self.meta_str = meta_str
        self.errors = errors

        self.coord = None
        self.coordsys = None
        self.coord_str = None
        self.type_ = type_
        self.region_type = region_type
        self.meta = copy.deepcopy(global_meta)
        self.shape = None
        self.include = include or '+'

        self.parse()

    def _raise_error(self, msg):
        if self.errors == 'warn':
            warn(msg, CRTFRegionParserWarning)
        elif self.errors == 'strict':
            raise CRTFRegionParserError(msg)

    def parse(self):
        """
        Starting point to parse the CRTF region string.
        """

        self.convert_meta()
        self.coordsys = self.meta.get('coord', 'image').lower()
        self.set_coordsys()
        self.convert_coordinates()
        self.make_shape()

    def set_coordsys(self):
        """
        Mapping to astropy's coordinate system name

        # TODO: needs expert attention (Most reference systems are not mapped)
        """
        if self.coordsys.lower() in self.coordsys_mapping:
            self.coordsys = self.coordsys_mapping[self.coordsys.lower()]

    def convert_coordinates(self):
        """
        Convert coordinate string to `~astropy.coordinates.Angle` or `~astropy.units.quantity.Quantity` objects
        """
        coord_list_str = regex_coordinate.findall(self.reg_str) + regex_length.findall(self.reg_str)
        coord_list = []

        if self.region_type == 'poly':
            if len(coord_list_str) < 4:
                self._raise_error('Not in proper format: {} polygon should have > 4 coordinates'.format(self.reg_str))
            if coord_list_str[0] != coord_list_str[-1]:
                self._raise_error("Not in proper format: '{0}', "
                                  "In polygon, the last and first coordinates should be same".format(self.reg_str))
        else:
            if len(coord_list_str) != len(self.language_spec[self.region_type]):
                self._raise_error("Not in proper format: '{0}', "
                                  "Does not contain expected number of parameters for the region '{1}'"
                                  .format(self.reg_str, self.region_type))

        for attr_spec, val_str in zip(self.language_spec[self.region_type], coord_list_str):

            if attr_spec == 'c':
                if len(val_str) == 2 and val_str[1] != '':
                    coord_list.append(CoordinateParser.parse_coordinate(val_str[0]))
                    coord_list.append(CoordinateParser.parse_coordinate(val_str[1]))
                else:
                    self._raise_error("Not in proper format: {0} should be a coordinate".format(val_str))
            if attr_spec == 'pl':
                if len(val_str) == 2 and val_str[1] != '':
                    coord_list.append(CoordinateParser.parse_angular_length_quantity(val_str[0]))
                    coord_list.append(CoordinateParser.parse_angular_length_quantity(val_str[1]))
                else:
                    self._raise_error("Not in proper format: {0} should be a pair of length".format(val_str))
            if attr_spec == 'l':
                if isinstance(val_str, six.string_types):
                    coord_list.append(CoordinateParser.parse_angular_length_quantity(val_str))
                else:
                    self._raise_error("Not in proper format: {0} should be a single length".format(val_str))
            if attr_spec == 's':
                if self.region_type == 'symbol':
                    if val_str in valid_symbols:
                        self.meta['symbol'] = val_str
                    else:
                        self._raise_error("Not in proper format: '{0}' should be a symbol".format(val_str))
                elif self.region_type == 'text':
                    self.meta['text'] = val_str[1:-1]

        self.coord = coord_list

    def convert_meta(self):
        """
        Parses the meta_str to python dictionary and stores in ``meta`` attribute.
        """
        if self.meta_str:
            self.meta_str = regex_meta.findall(self.meta_str + ',')
        if self.meta_str:
            for par in self.meta_str:
                if par[0] is not '':
                    val1 = par[0]
                    val2 = par[1]
                else:
                    val1 = par[2]
                    val2 = par[3]
                val1 = val1.strip()
                val2 = val2.strip()
                if val1 in CRTFParser.valid_global_keys or val1 == 'label':
                    if val1 in ('range', 'corr', 'labeloff'):
                        val2 = val2.split(',')
                        val2 = [x.strip() for x in val2]
                    self.meta[val1] = val2
                else:
                    self._raise_error("'{0}' is not a valid meta key".format(val1))

        self.meta['include'] = self.include != '-'
        self.include = self.meta['include']

        if 'range' in self.meta:
            self.meta['range'] = [u.Quantity(x) for x in self.meta['range']]

        self.meta['type'] = self.type_

    def make_shape(self):
        """
        Make shape object
        """
        if self.region_type == 'ellipse':
            self.coord[2:] = [x * 2 for x in self.coord[2:]]
            if len(self.coord) % 2 == 1:  # This checks if the angle is present.
                self.coord[-1] /= 2

        if self.region_type == 'box':
            x = (self.coord[0] + self.coord[2]) / 2
            y = (self.coord[1] + self.coord[3]) / 2
            w = u.Quantity(self.coord[0] - self.coord[2])
            h = u.Quantity(self.coord[1] - self.coord[3])
            self.coord = [x, y, abs(w), abs(h)]

        self.meta.pop('coord', None)

        self.shape = Shape(coordsys=self.coordsys,
                           region_type=reg_mapping['CRTF'][self.region_type],
                           coord=self.coord,
                           meta=self.meta,
                           composite=False,
                           include=self.include
                           )


class CoordinateParser(object):
    """
    Helper class to structure coordinate parser
    """
    @staticmethod
    def parse_coordinate(string_rep):
        """
        Parse a single coordinate
        """

        # Any CRTF coordinate representation (sexagesimal or degrees)
        if 'pix' in string_rep:
            return u.Quantity(string_rep[:-3], u.dimensionless_unscaled)
        if 'h' in string_rep or 'rad' in string_rep:
            return coordinates.Angle(string_rep)
        if len(string_rep.split('.')) >= 3:
            string_rep = string_rep.replace('.', ':', 2)

        return coordinates.Angle(string_rep, u.deg)

    @staticmethod
    def parse_angular_length_quantity(string_rep):
        """
        Given a string that is a number and a unit, return a
        Quantity of that string.Raise an Error If there is no unit.  e.g.:
            50" -> 50*u.arcsec
            50 -> CRTFRegionParserError : Units must be specified for 50
        """
        unit_mapping = {
            'deg': u.deg,
            'rad': u.rad,
            'arcmin': u.arcmin,
            'arcsec': u.arcsec,
            'pix': u.dimensionless_unscaled,
            '"': u.arcsec,
            "'": u.arcmin,
        }
        regex_str = re.compile(r'([0-9+,-.]*)(.*)')
        str = regex_str.search(string_rep)
        unit = str.group(2)
        if unit:
            if unit in unit_mapping:
                return u.Quantity(str.group(1), unit=unit_mapping[unit])
            return u.Quantity(str.group(1))
        else:
            raise CRTFRegionParserError('Units must be specified for {0} '.format(string_rep))



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np

from astropy import units as u
from astropy.coordinates.representation import UnitSphericalRepresentation

try:
    from astropy.coordinates.matrix_utilities import rotation_matrix as rotation_array
    def rotation_matrix(*args, **kwargs):
        return np.matrix(rotation_array(*args, **kwargs))
except ImportError:  # Astropy < 1.3
    from astropy.coordinates.angles import rotation_matrix

__all__ = ['rotate_polygon']


def rotate_polygon(lon, lat, lon0, lat0):
    """
    Given a polygon with vertices defined by (lon, lat), rotate the polygon
    such that the North pole of the spherical coordinates is now at (lon0,
    lat0). Therefore, to end up with a polygon centered on (lon0, lat0), the
    polygon should initially be drawn around the North pole.
    """

    # Create a representation object
    polygon = UnitSphericalRepresentation(lon=lon, lat=lat)

    # Determine rotation matrix to make it so that the circle is centered
    # on the correct longitude/latitude.
    m1 = rotation_matrix(-(0.5 * np.pi * u.radian - lat0), axis='y')
    m2 = rotation_matrix(-lon0, axis='z')
    transform_matrix = m2 * m1

    # Apply 3D rotation
    polygon = polygon.to_cartesian()
    polygon = polygon.transform(transform_matrix)
    polygon = UnitSphericalRepresentation.from_cartesian(polygon)

    return polygon.lon, polygon.lat
